%**********************************************************
% RALF - Reinforced Active Learning Formulation
%**********************************************************
% S. Ebert, M. Fritz, B. Schiele
% RALF: A Reinforced Active Learning Formulation for Object Class
% Recognition, CVPR 2012
%
% The authors and copyright holders are S. Ebert, M. Fritz, and B. Schiele.
% You might redistribute, reuse, and mofidy this code for research 
% purposes. Any commercial application needs a written agreement by one of
% the authors.

function obj = Sampling(obj)
% Time-varying sampling framework that combines exploration and
% exploitation criteria. If the option isRandomSampling is activated it
% samples the next sample with a uniform distribution.

%initialization with 1 label per class randomly drawn
if isempty(obj.traininds)
    for c=1:obj.numClasses
        cidx = find(obj.labels==c);
        perm = randperm(length(cidx));
        obj.traininds = [obj.traininds; cidx(perm(1))];
    end
else
    if ~obj.isRandomSampling
        %get scores for exploration
        exploration = str2func(obj.explore);
        obj = exploration(obj);
        %get scores for exploitation
        exploitation = str2func(obj.exploit);
        obj = exploitation(obj);
        
        %rank these scores
        if obj.isRanking
            rank(:,1) = GetRanking(obj.exploitVal);
            rank(:,2) = GetRanking(obj.exploreVal);
        else
            rank(:,1) = obj.exploitVal;
            rank(:,2) = obj.exploreVal;
        end
        %computes the final scores with trade-off parameter beta
        CE = obj.betas(obj.numTrain+1-obj.numClasses)*rank(:,1) + ...
            (1-obj.betas(obj.numTrain+1-obj.numClasses))*rank(:,2);
        %sort these scores
        [sortval sortidx] = sort(CE);
        %selection index
        id = 1;
        %if activated random selection with a Gamma distribution
        if obj.selRandom
            id = ceil(random('gam', 1, obj.gamB));
        end
        %add sample to the training set
        obj.traininds = [obj.traininds; obj.testinds(sortidx(id))];
        
        %reweighting of the node potential
        if strfind(char(obj.explore), 'NodePotential')
            obj.potBeta = 4/(1.25*obj.np_ra)^2;
            %all direct neighbors of the selected node are down weighted
            reweighting = obj.nodePotential - ...
                obj.nodePotential(obj.traininds(end)) * ...
                sum(exp(-obj.potBeta*obj.kernel.^2),2);
            %if new weights contains infinity or nan values nothing happens
            if sum(isnan(reweighting))==0 && sum(isinf(reweighting))==0
                obj.nodePotential = reweighting;
            end
        end
        %reweighting of the graph density
        if strfind(char(obj.explore), 'GraphDensityBased')
            idx = find(obj.Graph(obj.traininds(end),:)>0);
            reweighting = obj.graphDensity(idx) - ...
                obj.graphDensity(obj.traininds(end));
            %similar to node potential
            if sum(isnan(reweighting))==0 && sum(isinf(reweighting))==0
                obj.graphDensity(idx) = reweighting;
            end
        end
    else
        perm = randperm(length(obj.testinds));
        obj.traininds = [obj.traininds; obj.testinds(perm(1))];
    end
    
end

obj.testinds = setdiff(1:obj.numData, obj.traininds)';
obj.trainlabels = obj.labels(obj.traininds);
obj.testlabels = obj.labels(obj.testinds);
obj.numTrain = length(obj.traininds);