%**********************************************************
% RALF - Reinforced Active Learning Formulation
%**********************************************************
% S. Ebert, M. Fritz, B. Schiele
% RALF: A Reinforced Active Learning Formulation for Object Class
% Recognition, CVPR 2012
%
% The authors and copyright holders are S. Ebert, M. Fritz, and B. Schiele.
% You might redistribute, reuse, and mofidy this code for research 
% purposes. Any commercial application needs a written agreement by one of
% the authors.

tic;
initpath;
%% general parameter
%maximum number of runs per experiment
if ~isfield(obj, 'maxRun')
    obj.maxRun=5;
end

%fix seed at the beginning
if ~isfield(obj, 'seed')
    obj.seed = 1;
end
rand('twister', obj.seed);

%load data and labels in the respective variables; stored as single values
%for space issues
load(['data/' char(obj.dataset) '_hog.mat']);
%transform single values into double values as it is needed for the faster
%sparse representation (graph construction, confusion matrix,...)
obj.data = double(data);
obj.labels = double(labels);
clear data; clear labels;
%maximum number of classes
obj.numClasses = max(obj.labels);
%size of the dataset and dimensionality of the descriptor
[obj.numData obj.dim] = size(obj.data);

%% graph construction parameter
%distance measure for the graph construction
if ~isfield(obj, 'distance')
    obj.distance = 'L1';
end
%number of neighbors k for the k-NN graph
if ~exist('kLP', 'var')
    obj.kLP = 10;
end
%weighted graph with a Gaussian function (on/off); sigma is determined by a
%heuristic
if ~isfield(obj, 'isWeighted')
    obj.isWeighted = true;
end
%symmetric graph construction (on/off)
if(~isfield(obj, 'isSymmetric'))
    obj.isSymmetric = true;
end
%graph construction
obj = ConstructkNNGraph(obj);

%% Label propagation parameter 
%algorithm; default LPZhou [ZhouNIPS2004];
%other options: KNN, SVM (requires libSVM), LPZhu [ZhuICML2003]
if ~isfield(obj, 'algo')
    obj.algo = str2func('LPZhou');
end
%regularization parameter alpha in (0,1]
%the larger alpha the larger the overwriting of the original labels
if ~isfield(obj, 'alphaLP')
    obj.alphaLP = 0.8;
end
%maximum number of iteration for the iterative propagation solution
%should be small to avoid overfitting and to trust more in strong nearest
%neighbors
if(~isfield(obj, 'maxIterLP'))
    obj.maxIterLP = 10;
end

%% Active Learning parameter
% transforms the scores (entropy,...) into ranked values
if ~isfield(obj, 'isRanking')
    obj.isRanking = true;
end
% random selection of the next queried sample with a gamma distribution
% default: query always the next labels with the smallest score
if ~isfield(obj,'selRandom')
    obj.selRandom = false;
end
%
if ~isfield(obj, 'gamB')
    obj.gamB = 5;
end
%number of requested labels per class in average
if ~isfield(obj, 'numPerClass')
    obj.numPerClass = 5;
end
%number of active selections with at least 100 queries (important only for
%ETH otherwise there are only 40 labels)
obj.numtrain = max(100, 5*obj.numClasses);
%default exploration criteria
if ~isfield(obj, 'explore')
    obj.explore = 'GraphDensityBasedRep';
end
%default exploitation criteria
if ~isfield(obj, 'exploit')
    obj.exploit = 'MarginBasedInf';
end
%random sampling baseline; sampling with a uniform distribution
if ~isfield(obj, 'isRandomSampling')
    obj.isRandomSampling = false;
end
%time-varying trade-off function
if ~isfield(obj, 'betaFun')
    obj.betaFun = 'log(x)';
end
%calculates the beta values for a time-varying trade-off between
%exploration and exploitation; default: log(x) scaled into [0,1]
GetBetaValues;

%% exploration criteria are pre-computed
% 1) node potential 
%see [3] Cebron @ DMKD 2009
%parameter from the paper
obj.np_ra = 0.4;
obj.np_Alpha = 4/(obj.np_ra^2);
%euclidean distance that is also used the kernel farthest first criteria
if ~isfield(obj, 'kernel')
    obj.kernel = obj.data*obj.data';
    obj.kernel = sqrt(bsxfun(@plus, diag(obj.kernel), diag(obj.kernel)')-2*obj.kernel);
end
%calculates node potential
obj.nodePotential = sum(exp(-obj.np_Alpha*obj.kernel.^2),2);

% 3) normalized graph density: sum of edges divided by number of edges
obj.graphDensity = full(sum(obj.Graph,2)./sum(obj.Graph>0,2));

fprintf('Initialization done after %2.1f seconds.\n',toc);

%% beta estimation parameter
% threshold to have a minimum of exploitation/exploration
obj.betaEstEps = 0.05;

% learning rate
obj.betaEstLambda = 0.5;

%% Q learning parameter
% learning rate controls the influence of the current reward 
obj.Qalpha = 0.5;

% discount factor weights the future reward; gamma in (0,1); if
% gamma=0 only the current reward is considered
obj.Qgamma = 1;